######################################################################
######################################################################

#
# getvisual:
#
# return the OpenGL compatible visual with the greatest depth
#

proc getvisual {which} {
  switch $which {

    canvas {
      set visuals [sminfo visuals canvas];
    }

    widget {
      set visuals [sminfo visuals widget];
    }

    default {
      error "bad option \"$which\": must be canvas or widget";
    }
  }

  set mydepth 0;
  set myvisual {};

  foreach visual $visuals {
    set depth [lindex $visual 2];
    if {$mydepth < $depth} {
      set mydepth $depth;
      set myvisual [lrange $visual 1 2];
    }
  }

  return $myvisual;
}


######################################################################
######################################################################

#
# normalize_vector:
#
# normalize a vector
#

proc normalize_vector {x y z} {
  set d [expr sqrt($x * $x + $y * $y + $z * $z)];
  if {$d > 0} {
    set d [expr 1.0 / $d];
    list [expr $x * $d] [expr $y * $d] [expr $z * $d];
  } else {
    return {};
  }
}


######################################################################
######################################################################

#
# dot_product:
#
# compute the dot product of two vectors
#

proc dot_product {ax ay az bx by bz} {
  expr $ax * $bx + $ay * $by + $az * $bz;
}


######################################################################
######################################################################

#
# cross_product:
#
# compute the cross product of two vectors
#

proc cross_product {ax ay az bx by bz} {
  list [expr $ay * $bz - $az * $by] \
       [expr $az * $bx - $ax * $bz] \
       [expr $ax * $by - $ay * $bx];
}


######################################################################
######################################################################

#
# rotate_about_x:
#
# rotate a vector about the x-axis
#

proc rotate_about_x {x y z angle} {

  set sine [expr sin($angle)];
  set cosine [expr cos($angle)];

  return [list $x \
	       [expr $cosine * $y - $sine * $z] \
	       [expr $sine * $y + $cosine * $z] \
	  ];
}


######################################################################
######################################################################

#
# rotate_about_y:
#
# rotate a vector about the y-axis
#

proc rotate_about_y {x y z angle} {

  set sine [expr sin($angle)];
  set cosine [expr cos($angle)];

  return [list [expr $cosine * $x + $sine * $z] \
	       $y \
	       [expr -$sine * $x + $cosine * $z] \
	  ];
}


######################################################################
######################################################################

#
# rotate_about_z:
#
# rotate a vector about the z-axis
#

proc rotate_about_z {x y z angle} {

  set sine [expr sin($angle)];
  set cosine [expr cos($angle)];

  return [list [expr $cosine * $x - $sine * $y] \
	       [expr $sine * $x + $cosine * $y] \
	       $z \
	  ];
}


######################################################################
######################################################################

proc rotate {x y z m0 m1 m2 m3 m4 m5 m6 m7 m8} {

#
# rotate:
#
# rotate a vector using a pre-computed rotation matrix
#

  return [list [expr $x * $m0 + $y * $m1 + $z * $m2] \
	       [expr $x * $m3 + $y * $m4 + $z * $m5] \
	       [expr $x * $m6 + $y * $m7 + $z * $m8] \
	       ];
}


######################################################################
######################################################################

#
# rotation_axis:
#
# compute the matrix for rotation about an arbitrary axis
#

proc rotation_axis {cx cy cz angle} {

  set sine [expr sin($angle)];
  set cosine [expr cos($angle)];

  set m00 [expr $cx * $cx + $cosine * (1 - $cx * $cx)];
  set m01 [expr $cx * $cy * (1 - $cosine) - $cz * $sine];
  set m02 [expr $cz * $cx * (1 - $cosine) + $cy * $sine];

  set m10 [expr $cx * $cy * (1 - $cosine) + $cz * $sine];
  set m11 [expr $cy * $cy + $cosine * (1 - $cy * $cy)];
  set m12 [expr $cy * $cz * (1 - $cosine) - $cx * $sine];

  set m20 [expr $cz * $cx * (1 - $cosine) - $cy * $sine];
  set m21 [expr $cy * $cz * (1 - $cosine) + $cx * $sine];
  set m22 [expr $cz * $cz + $cosine * (1 - $cz * $cz)];

  list $m00 $m01 $m02 $m10 $m11 $m12 $m20 $m21 $m22;
}


######################################################################
######################################################################

#
# rotation_vectors:
#
# compute the rotation matrix about an axis in the direction of the
# cross-product of two vectors
#

proc rotation_vectors {u0 u1 u2 v0 v1 v2} {

  set axis [eval [concat normalize_vector [cross_product $u0 $u1 $u2 $v0 $v1 $v2]]];
  if {$axis == {}} {
    return {};
  }

  set cosine [dot_product $u0 $u1 $u2 $v0 $v1 $v2];
  set sine [expr sqrt(1 - $cosine * $cosine)];

  set cx [lindex $axis 0];
  set cy [lindex $axis 1];
  set cz [lindex $axis 2];

  set m00 [expr $cx * $cx + $cosine * (1 - $cx * $cx)];
  set m01 [expr $cx * $cy * (1 - $cosine) - $cz * $sine];
  set m02 [expr $cz * $cx * (1 - $cosine) + $cy * $sine];

  set m10 [expr $cx * $cy * (1 - $cosine) + $cz * $sine];
  set m11 [expr $cy * $cy + $cosine * (1 - $cy * $cy)];
  set m12 [expr $cy * $cz * (1 - $cosine) - $cx * $sine];

  set m20 [expr $cz * $cx * (1 - $cosine) - $cy * $sine];
  set m21 [expr $cy * $cz * (1 - $cosine) + $cx * $sine];
  set m22 [expr $cz * $cz + $cosine * (1 - $cz * $cz)];

  return [list $m00 $m01 $m02 $m10 $m11 $m12 $m20 $m21 $m22];
}


######################################################################
######################################################################

#
# compute_angle_2D:
#
# compute the angle a line drawn from the center of the window <w>
# to the point (x,y) makes with the horizontal.
#

proc compute_angle_2D {w x y} {

  global PI PI_2;

  set width [winfo width $w];
  set height [winfo height $w];

  if {($width == -1) || ($height == -1)} {
    return 0;
  }

  set dx [expr $x - $width * 0.5];
  set dy [expr $y - $height * 0.5];

  if {$dx == 0} {
    if {$dy == 0} {
      return 0;
    } elseif {$dy > 0} {
      return $PI_2;
    } else {
      return -$PI_2;
    }
  } else {
    set angle [expr atan($dy / $dx)];

    if {$dx >= 0} {
      return $angle;
    } elseif {$dy >= 0} {
      return [expr $angle + $PI];
    } else {
      return [expr $angle - $PI];
    }
  }
}


######################################################################
######################################################################

proc compute_vector_3D {w x y maxx maxy} {

#
# compute_vector_3D:
#
# compute a vector from the cursor position (given by x,y) to be used as the axis
# of rotation.
#
  set rx [expr double($x) / $maxx];
  set ry [expr double($y) / $maxy];

  set d [expr $rx * $rx + $ry * $ry];
  if {$d >= 1} {
    set f [expr 1.0 / sqrt($d)];
    set vector [list [expr $rx * $f] [expr $ry * $f] 0];
  } else {
    set vector [list $rx $ry [expr sqrt(1.0 - $d)]];
  }
  return $vector;
}
