/********************************************************************************
 *
 * wwwanchor.c
 *
 ********************************************************************************/

#include <stdlib.h>
#include <tk.h>

#include "vrml.h"


/********************************************************************************
 *
 * Vrml_read_WWWAnchor
 *
 * Read in a WWWAnchor  node.
 *
 * fields:	name		""	SFString
 *		description	""	SFString
 *		map		NONE	SFEnum
 *
 *******************************************************************************/

Node *
Vrml_read_WWWAnchor(interp, channel, argv, node, names, inlines, textures)
     Tcl_Interp *interp;
     Tcl_Channel channel;
     char *argv;
     Node *node;
     Node **names, **inlines, **textures;
{
    int dofield, nchild;
    char *field;
    Node *child, *last;
    NodeWWWAnchor *anchor;

    anchor = &node->node.wwwAnchor;
    anchor->name = anchor->description = NULL;
    anchor->map = ENUM_NONE;

    /* get open curly bracket */
    switch(Vrml_get_token(channel, &field)) {
	
      case TOKEN_OUT_OF_MEMORY:
	Tcl_AppendResult(interp, argv, ": out of memory while reading WWWAnchor", (char *) NULL);
	return (Node *) -1;
	
      case TOKEN_EOF:
      case TOKEN_END:
	Tcl_AppendResult(interp, argv, ": unexpected end of input while reading WWWAnchor", (char *) NULL);
	return (Node *) -1;

      case TOKEN_OPEN_CURLY:
	break;
	
      case TOKEN_WORD:
	Vrml_free_token(field);
	
      default:
	Tcl_AppendResult(interp, argv, ": bad WWWAnchor format", (char *) NULL);
	return (Node *) -1;
    }

    /* parse all fields until first child node or close curly bracket */
    dofield = 1;
    while (dofield) {
	
	switch(Vrml_get_token(channel, &field)) {
	    
	  case TOKEN_OUT_OF_MEMORY:
	    Tcl_AppendResult(interp, argv, ": out of memory while reading WWWAnchor", (char *) NULL);
	    return (Node *) -1;

	  case TOKEN_EOF:
	  case TOKEN_END:
	    Tcl_AppendResult(interp, argv, ": unexpected end of input while reading WWWAnchor", (char *) NULL);
	    return (Node *) -1;
	    
	  case TOKEN_CLOSE_CURLY:
	    return node;
	    
	  case TOKEN_WORD:
	    break;
	    
	  default:
	    Tcl_AppendResult(interp, argv, ": bad WWWAnchor format", (char *) NULL);
	    return (Node *) -1;
	}

	if (!strcmp(field, "name")) {
	    if (anchor->name) {
		Vrml_free_token(anchor->name);
		anchor->name = NULL;
	    }
	    if (Vrml_read_SFString(interp, channel, argv, &anchor->name, NULL, NULL) != TCL_OK) {
		goto err;
	    }
	}
	else if (!strcmp(field, "description")) {
	    if (anchor->description) {
		Vrml_free_token(anchor->description);
		anchor->description = NULL;
	    }
	    if (Vrml_read_SFString(interp, channel, argv, &anchor->description, NULL, NULL) != TCL_OK) {
		goto err;
	    }
	}
	else if (!strcmp(field, "map")) {
	    if (Vrml_read_SFEnum(interp, channel, argv, (unsigned long) (ENUM_NONE | ENUM_POINT), &anchor->map) != TCL_OK) {
		goto err;
	    }
	}
	else {
	    dofield = 0;
	}
	if (dofield) {
	    Vrml_free_token(field);
	}
    }

    /* parse child nodes */
    last = NULL;
    nchild = 0;
    child = Vrml_read_node(interp, channel, argv, names, inlines, textures, 1, field);

    Vrml_free_token(field);
    field = NULL;

    while ((child != NULL) && (child != (Node *) -1)) {
	nchild++;
	if (last) {
	    last->sibling = child;
	}
	else {
	    node->child = child;
	}
	last = child;
	child = Vrml_read_node(interp, channel, argv, names, inlines, textures, 1, NULL);
    }
    if (child == NULL) {
	return node;
    }
  
  err:
    if (field) Vrml_free_token(field);
    Vrml_free_WWWAnchor(node);
    return (Node *) -1;
}


/********************************************************************************
 *
 * Vrml_free_WWWAnchor
 *
 * Free an WWWAnchor node.
 *
 * fields:	name		""	SFString
 *		description	""	SFString
 *		map		NONE	SFEnum
 *
 *******************************************************************************/

void
Vrml_free_WWWAnchor(node)
     Node *node;
{
    Node *child, *sibling;
    NodeWWWAnchor *anchor = &node->node.wwwAnchor;
    
    if (anchor->name) Vrml_free_token(anchor->name);
    if (anchor->description) Vrml_free_token( anchor->description);
    for (child = node->child; child; child = sibling) {
	sibling = child->sibling;
	Vrml_free_node(child);
    }
}


/********************************************************************************
 *
 * Vrml_render_WWWAnchor
 *
 * Render a WWWAnchor node.
 *
 * fields:	name		""	SFString
 *		description	""	SFString
 *		map		NONE	SFEnum
 *
 *******************************************************************************/

int
Vrml_render_WWWAnchor(interp, node, state)
     Tcl_Interp *interp;
     Node *node;
     State *state;
{
    SFVec3f pos, fwd, up, scale;

    node->objects = state->objects;
    state->objects = node;

    Vrml_object_state(state, &pos, &fwd, &up, &scale);

    node->anchors = state->anchors;
    state->anchors = node;

    if ((Vrml_new_object(interp, state, node, "polygon", &pos, &fwd, &up, 0, NULL) != TCL_OK) ||
	(Vrml_push_model(interp, state, node) != TCL_OK) ||
	(Vrml_push_transform(interp, state) != TCL_OK) ||
	(Vrml_push_surface(interp, state) != TCL_OK)) {
	return 2;
    }

    return 0;
}


/********************************************************************************
 *
 * Vrml_finish_WWWAnchor
 *
 * Finish rendering a WWWAnchor node.
 *
 *******************************************************************************/

int
Vrml_finish_WWWAnchor(interp, node, state)
     Tcl_Interp *interp;
     Node *node;
     State *state;
{
    Vrml_pop_model(state);
    Vrml_pop_transform(state);
    Vrml_pop_surface(state);

    return 0;
}
